//   ( STL ,  )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<stdlib.h>
#include<assert.h>


//!  
struct node_st{
	node_st(int);								
	int value;									//!<  
	node_st *next;
	node_st *child;
	static node_st *Merge(node_st*,node_st*);	//!<   
	void Clear();								//!<  
	void Dump(int depth);						//!<  
	int GetCount();								//!<    (  )
};

//!   
struct item_st{
	int h;
	node_st *node;
	item_st *next;
	item_st(int h,node_st*);
};



//!  
class BiHeap{
	item_st *list;								//!<   
	void Insert(int h,node_st*);				//!<     
public:
	BiHeap();
	~BiHeap();
	void Insert(int v);							//!<    
	bool isEmpty();								//!<   
	int  Pop();									//!<   
	void Dump();								//!<  
};


item_st::item_st(int h_,node_st *node_)
{
	h=h_;
	node=node_;
}



//! \param v  
node_st::node_st(int v)
{
	value=v;
	next=child=0;
}


int node_st::GetCount()
{
	int count=0;
	node_st *temp=this;
	while(temp) count++,temp=temp->next;
	return count;
}

//!   ,      
node_st *node_st::Merge(node_st *node1,node_st *node2)
{
	node_st *node;
	assert(!node1->next && !node2->next);
	assert(node1->GetCount()==node2->GetCount());
	if(node1->value<node2->value) {
		node=node1;
		node2->next=node->child;
		node->child=node2;
	}else{
		node=node2;
		node1->next=node->child;
		node->child=node1;
	}
	return node;
}

void node_st::Clear()
{
	if(next) next->Clear();
	if(child) child->Clear();
	delete this;
}

//! \param depth  
void node_st::Dump(int depth)
{
	for(int n=0; n<depth; n++) putchar(' ');
	printf("[%d]\n",value);
	node_st *temp=child;
	while(temp) {
		temp->Dump(depth+2);
		temp=temp->next;
	}
}


BiHeap::BiHeap()
{
	list=0;
}


BiHeap::~BiHeap()
{
	item_st *next;
	while(list) {
		next=list->next;
		list->node->Clear();
		delete list;
		list=next;
	}
}


//! \param v  
void BiHeap::Insert(int v)
{
	Insert(0,new node_st(v));
}

//! \param  
void BiHeap::Insert(int h,node_st *add)
{
	add->next=0;
	item_st **old,*item;
	for(;;) {
		old=&list;
		//    
		for(;;) {
			item=*old;
			if(!item) break;
			if(item->h==h) break;
			old=&item->next;
		}
		if(!item) break;
		//    
		add=node_st::Merge(add,item->node); h++;
		//     
		*old=item->next; delete item;
	}
	//      
	item=new item_st(h,add);
	item->next=list; list=item;
}


int BiHeap::Pop()
{
	int min,h;
	node_st *node,*next;
	item_st **old,*item,**old_res,*temp; 
	min=list->node->value+1;
	old=&list;
	//     
	for(;;) {
		temp=*old; if(!temp) break;
		if(min>temp->node->value) item=temp,old_res=old,min=temp->node->value;
		old=&temp->next;
	}
	h=item->h;
	node=item->node->child;
	//    
	*old_res=item->next;
	delete item->node;
	delete item;
	//    
	while(node) 
	{
		assert(node->GetCount()==h);
		h--;
		next=node->next;
		Insert(h,node);
		node=next;
	}
	return min;
}


bool BiHeap::isEmpty()
{
	return !list;
}

void BiHeap::Dump()
{
	item_st *temp=list;
	printf("[dump]\n");
	while(temp) temp->node->Dump(0),temp=temp->next;
}



#define SIZE 30

void main()
{
	BiHeap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	while(!heap.isEmpty()) printf("(%d)",heap.Pop());
	putchar('\n');
	printf("[press enter]\n");
	getchar();
}

